@@ -17,7 +17,7 @@ use rand::seq::SliceRandom;
 use rand::SeedableRng;
 use rayon::prelude::*;
 use regex::RegexSet;
-use std::collections::HashSet;
+use std::collections::{HashMap, HashSet};
 use std::fs::File;
 use std::io::prelude::*;
 use std::io::{BufReader, BufWriter};
@@ -26,32 +26,7 @@ use std::process::{Command, Stdio};
 use std::sync::mpsc::{channel, Receiver};
 use std::sync::{Arc, Mutex};
 use std::time::Duration;
-use std::{fmt, time::Instant};
-
-// Wrapper for displaying a duration in h:m:s (integer seconds, rounded down)
-struct HMSDuration(Duration);
-impl fmt::Display for HMSDuration {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let mut secs = self.0.as_secs();
-        let mut has_hours = false;
-        if secs >= 3600 {
-            write!(f, "{}:", secs / 3600)?;
-            secs %= 3600;
-            has_hours = true;
-        }
-        if secs >= 60 {
-            if has_hours {
-                write!(f, "{:02}:", secs / 60)?;
-            } else {
-                write!(f, "{}:", secs / 60)?;
-            }
-            secs %= 60;
-            write!(f, "{:02}", secs)
-        } else {
-            write!(f, "{}", secs)
-        }
-    }
-}
+use std::time::Instant;
 
 pub trait Deqp {
     fn skips(&self) -> Option<&RegexSet>;
@@ -66,39 +41,8 @@ pub trait Deqp {
         "".to_string()
     }
 
-    fn status_update(&self, results: &RunnerResults, total_tests: u32) {
-        let duration = results.time.elapsed();
-
-        print!(
-            "{}, Duration: {duration}",
-            results.result_counts,
-            duration = HMSDuration(duration),
-        );
-
-        // If we have some tests completed, use that to estimate remaining runtime.
-        let duration = duration.as_secs_f32();
-        if results.result_counts.total != 0 {
-            let average_test_time = duration / results.result_counts.total as f32;
-            let remaining = average_test_time * (total_tests - results.result_counts.total) as f32;
-
-            print!(
-                ", Remaining: {}",
-                HMSDuration(Duration::from_secs_f32(remaining))
-            );
-        }
-
-        println!();
-    }
-
     fn baseline(&self) -> &RunnerResults;
 
-    fn min_tests_per_group(&self) -> usize {
-        20
-    }
-    fn tests_per_group(&self) -> usize {
-        500
-    }
-
     fn qpa_to_xml(&self) -> Option<&PathBuf> {
         None
     }
@@ -133,43 +77,6 @@ pub trait Deqp {
         status
     }
 
-    fn results_collection(
-        &self,
-        run_results: &mut RunnerResults,
-        total_tests: u32,
-        receiver: Receiver<Result<Vec<RunnerResult>>>,
-    ) {
-        let update_interval = Duration::new(2, 0);
-
-        self.status_update(run_results, total_tests);
-        let mut last_status_update = Instant::now();
-
-        for group_results in receiver {
-            match group_results {
-                Ok(group_results) => {
-                    for result in group_results {
-                        if run_results.tests.contains_key(&result.test) {
-                            error!("Duplicate test result for {}", &result.test);
-                        }
-                        assert!(!run_results.tests.contains_key(&result.test));
-
-                        run_results.record_result(result);
-                    }
-                }
-                Err(e) => {
-                    println!("Error: {}", e);
-                }
-            }
-            if last_status_update.elapsed() >= update_interval {
-                self.status_update(run_results, total_tests);
-                last_status_update = Instant::now();
-            }
-        }
-
-        // Always print the final results
-        self.status_update(run_results, total_tests);
-    }
-
     fn skip_test(&self, test: &str) -> bool {
         if let Some(skips) = self.skips() {
             skips.is_match(test)
@@ -223,7 +130,7 @@ pub trait Deqp {
         let mut remaining_tests = Vec::new();
         for test in tests {
             let test = test.as_ref().clone();
-            if self.skip_test(&test.name()) {
+            if self.skip_test(format!("{}{}", self.prefix(), &test.name()).as_str()) {
                 caselist_results.push(RunnerResult {
                     test: test.name().to_owned(),
                     status: RunnerStatus::Skip,
@@ -249,9 +156,9 @@ pub trait Deqp {
                         /* Remove the reported test from our list of tests to run.  If it's not in our list, then it's
                          * a subtest.
                          */
-                        if let Some(position) =
-                            remaining_tests.iter().position(|x| x.name() == result.test)
-                        {
+                        if let Some(position) = remaining_tests.iter().position(|x| {
+                            x.name() == result.test.strip_prefix(self.prefix()).unwrap()
+                        }) {
                             remaining_tests.swap_remove(position);
                         } else if !result.subtest {
                             error!(
@@ -286,27 +193,34 @@ pub trait Deqp {
         Ok(caselist_results)
     }
 
-    fn split_tests_to_groups(&self, mut tests: Vec<TestCase>) -> Vec<(u32, Vec<TestCase>)> {
+    fn split_tests_to_groups(
+        &self,
+        mut tests: Vec<TestCase>,
+        tests_per_group: usize,
+        min_tests_per_group: usize,
+    ) -> Vec<(&Self, Vec<TestCase>)> {
         // Shuffle the test groups using a deterministic RNG so that every run gets the same shuffle.
         tests.shuffle(&mut StdRng::from_seed([0x3bu8; 32]));
 
         // Make test groups of tests_per_group() (512) tests, or if
         // min_tests_per_group() is lower than that, then 1/32nd of the
         // remaining tests down to that limit.
-        let mut test_groups: Vec<(u32, Vec<TestCase>)> = Vec::new();
+        let mut test_groups: Vec<(&Self, Vec<TestCase>)> = Vec::new();
         let mut remaining = tests.len();
-        let mut i = 0u32;
         while remaining != 0 {
-            let min = usize::min(self.min_tests_per_group(), remaining);
-            let group_len = usize::min(usize::max(remaining / 32, min), self.tests_per_group());
+            let min = usize::min(min_tests_per_group, remaining);
+            let group_len = usize::min(usize::max(remaining / 32, min), tests_per_group);
             remaining -= group_len;
 
-            test_groups.push((i, tests.split_off(remaining)));
-            i += 1;
+            test_groups.push((self, tests.split_off(remaining)));
         }
 
         test_groups
     }
+
+    fn prefix(&self) -> &str {
+        ""
+    }
 }
 
 pub struct DeqpCommand {
@@ -318,8 +232,8 @@ pub struct DeqpCommand {
     pub qpa_to_xml: Option<PathBuf>,
     pub baseline: RunnerResults,
     pub timeout: Duration,
-    pub tests_per_group: usize,
-    pub min_tests_per_group: usize,
+    pub env: HashMap<String, String>,
+    pub prefix: String,
 }
 
 pub struct PiglitCommand {
@@ -541,6 +455,7 @@ impl Deqp for DeqpCommand {
             // otherwise fills the logs with warnings about GL errors that are
             // thrown (you're running deqp!  Of course it makes GL errors!)
             .env("MESA_DEBUG", "silent")
+            .envs(self.env.iter())
             .spawn()
             .with_context(|| format!("Failed to spawn {}", &self.deqp.display()))?;
 
@@ -556,6 +471,10 @@ impl Deqp for DeqpCommand {
 
         let mut deqp_results = deqp_results.context("parsing results")?;
 
+        for result in &mut deqp_results {
+            result.name = format!("{}{}", self.prefix, result.name);
+        }
+
         let stderr = BufReader::new(child.stderr.as_mut().context("opening stderr")?);
         for line in stderr.lines().flatten() {
             // If the driver has ASan enabled and it detected leaks, then mark
@@ -633,17 +552,13 @@ impl Deqp for DeqpCommand {
         &self.baseline
     }
 
-    fn tests_per_group(&self) -> usize {
-        self.tests_per_group
-    }
-
-    fn min_tests_per_group(&self) -> usize {
-        self.min_tests_per_group
-    }
-
     fn qpa_to_xml(&self) -> Option<&PathBuf> {
         self.qpa_to_xml.as_ref()
     }
+
+    fn prefix(&self) -> &str {
+        &self.prefix
+    }
 }
 
 impl Deqp for PiglitCommand {
@@ -660,7 +575,7 @@ impl Deqp for PiglitCommand {
         // We only run one piglit command in a test group.  This means that
         // flake detection doesn't have to run irrelevant tests, and makes our
         // log file handling easier.
-        assert!(self.tests_per_group() == 1);
+        assert_eq!(tests.len(), 1);
         let test = &tests[0];
 
         let test = match test {
@@ -838,21 +753,64 @@ impl Deqp for PiglitCommand {
     fn baseline(&self) -> &RunnerResults {
         &self.baseline
     }
+}
 
-    fn tests_per_group(&self) -> usize {
-        1
+fn results_collection<W: Write>(
+    status_output: &mut W,
+    run_results: &mut RunnerResults,
+    total_tests: u32,
+    receiver: Receiver<Result<Vec<RunnerResult>>>,
+) {
+    let update_interval = Duration::new(2, 0);
+
+    run_results.status_update(status_output, total_tests);
+    let mut last_status_update = Instant::now();
+
+    for group_results in receiver {
+        match group_results {
+            Ok(group_results) => {
+                for result in group_results {
+                    if run_results.tests.contains_key(&result.test) {
+                        error!(
+                            "Duplicate test result for {}, marking test failed",
+                            &result.test
+                        );
+                        let mut fail_result = result;
+                        fail_result.status = RunnerStatus::Fail;
+                        run_results.record_result(fail_result);
+                    } else {
+                        run_results.record_result(result);
+                    }
+                }
+            }
+            Err(e) => {
+                println!("Error: {}", e);
+            }
+        }
+        if last_status_update.elapsed() >= update_interval {
+            run_results.status_update(status_output, total_tests);
+            last_status_update = Instant::now();
+        }
     }
+
+    // Always print the final results
+    run_results.status_update(status_output, total_tests);
 }
 
 // Splits the list of tests to groups and parallelize them across all cores, collecting results in
 // a separate thread
-pub fn parallel_deqp<D>(deqp: &D, tests: Vec<TestCase>) -> Result<RunnerResults>
+pub fn parallel_deqp<D, W>(
+    status_output: W,
+    test_groups: Vec<(&D, Vec<TestCase>)>,
+) -> Result<RunnerResults>
 where
     D: Deqp,
     D: Sync,
+    W: Write,
+    W: Sync,
+    W: Send,
 {
-    let test_count = tests.len();
-    let test_groups = deqp.split_tests_to_groups(tests);
+    let test_count = test_groups.iter().map(|x| x.1.len() as u32).sum();
 
     let mut run_results = RunnerResults::new();
 
@@ -860,18 +818,21 @@ where
     // printing the console status output but also computing the run_results.
     let (sender, receiver) = channel::<Result<Vec<RunnerResult>>>();
 
+    let mut status_output = status_output;
+
     crossbeam_utils::thread::scope(|s| {
         // Spawn the results collection in a crossbeam scope, so that it doesn't
         // take a slot in rayon's thread pool.
-        s.spawn(|_| deqp.results_collection(&mut run_results, test_count as u32, receiver));
+        s.spawn(|_| results_collection(&mut status_output, &mut run_results, test_count, receiver));
 
         // Rayon parallel iterator takes our vector and runs it on its thread
         // pool.
         test_groups
             .into_iter()
+            .enumerate()
             .par_bridge()
-            .try_for_each_with(sender, |sender, (i, tests)| {
-                sender.send(deqp.process_caselist(tests, i))
+            .try_for_each_with(sender, |sender, (i, (deqp, tests))| {
+                sender.send(deqp.process_caselist(tests, i as u32))
             })
             .unwrap();
 
@@ -992,18 +953,6 @@ mod tests {
         assert_eq!(results.is_success(), false);
     }
 
-    #[test]
-    fn hms_display() {
-        assert_eq!(format!("{}", HMSDuration(Duration::new(15, 20))), "15");
-        assert_eq!(format!("{}", HMSDuration(Duration::new(0, 20))), "0");
-        assert_eq!(format!("{}", HMSDuration(Duration::new(70, 20))), "1:10");
-        assert_eq!(format!("{}", HMSDuration(Duration::new(69, 20))), "1:09");
-        assert_eq!(
-            format!("{}", HMSDuration(Duration::new(3735, 20))),
-            "1:02:15"
-        );
-    }
-
     #[test]
     fn filter_qpa_success() {
         assert_eq!(
